#ifndef GEOMETRY_UTILS_H
#define GEOMETRY_UTILS_H

class Vector3f;

#include <math/Random.h>
#include <vecmath/Vector4f.h>
#include <geometry/BoundingBox2f.h>
#include <geometry/BoundingBox3f.h>

class GeometryUtils
{
public:

	// TODO: static const float epsilon
	// have most functions take in an optional epsilon
	static float EPSILON;

	static BoundingBox2f triangleBoundingBox( const Vector2f& v0, const Vector2f& v1, const Vector2f& v2 );

	static Vector2f triangleCentroid( const Vector2f& v0, const Vector2f& v1, const Vector2f& v2 );

	// returns true if two points (p0, p1) are on the same side of
	// the line determined by (v0, v1)
	static bool pointsOnSameSide( const Vector2f& p0, const Vector2f& p1,
		const Vector2f& v0, const Vector2f& v1 );

	static bool pointInTriangle( const Vector2f& point,
		const Vector2f& v0, const Vector2f& v1, const Vector2f& v2 );

	// given a point p inside a triangle and the triangle's 3 vertices v_i
	// returns its barycentric coordinates b = [ l0, l1, l2 ]
	// p = l0 * v0 + l1 * v1 + l2 * v2, where l2 = 1 - l0 - l1
	static Vector3f euclideanToBarycentric( const Vector2f& p,
		const Vector2f& v0, const Vector2f& v1, const Vector2f& v2 );

	// given barycentric coordinates b = [ b0, b1, b2 ]
	// returns its euclidean coordinates p = b0 * v0 + b1 * v1 + b2 * v2
	static Vector2f barycentricToEuclidean( const Vector3f& b,
		const Vector2f& v0, const Vector2f& v1, const Vector2f& v2 );

	static bool pointInBox( const Vector3f& crPoint, const BoundingBox3f& bbox );
	
	static bool pointInsideSphere( const Vector3f& crPoint,
		const Vector3f& crSphereCenter, float sphereRadius );

	// plane is defined by dot( plane.xyz, X ) = plane.w
	static bool rayPlaneIntersection( const Vector3f& crRayOrigin, const Vector3f& crRayDirection,
		const Vector4f& crPlane,
		Vector3f& rIntersectionPoint );

	// ray triangle intersection
	// returns true if the ray intersects the triangle
	// TODO: one sided intersection is also in there, ifdefed
	static bool rayTriangleIntersection( const Vector3f& crRayOrigin, const Vector3f& crRayDirection,
		const Vector3f& crV0, const Vector3f& crV1, const Vector3f& crV2,
		float* t, float* u, float* v );

	static bool triangleAABBOverlap( Vector3f* pv0, Vector3f* pv1, Vector3f* pv2,
		BoundingBox3f* pBox );

	// given u along edge 0 -> 1 and v along edge 0 -> 2
	// and the interpolants on the vertices
	// returns the interpolated value
	static float triangleInterpolation( float interpolant0, float interpolant1, float interpolant2,
		float u, float v );

	// plane is defined by dot( plane.xyz, X ) = plane.w
	static float pointToPlaneDistance( const Vector3f& point, const Vector4f& plane );

#if 0
	// TODO: make this non-nasty
	static void tripleSphereIntersection( Vector3f* c0, float r0,
		Vector3f* c1, float r1,
		Vector3f* c2, float r2,
		int* numIntersections,
		Vector3f* intersect0,
		Vector3f* intersect1 );
#endif

	static Vector3f randomPointInSphere( float radius, const Random& random );

private:

};

#endif
