#ifndef MATRIX3F_H
#define MATRIX3F_H

#include <cstdio>

class Matrix2f;
class Quat4f;
class Vector3f;

// 3x3 Matrix, stored in column major order (OpenGL style)
class Matrix3f
{
public:

	Matrix3f();
	Matrix3f( float m00, float m01, float m02,
		float m10, float m11, float m12,
		float m20, float m21, float m22 );

	// bCols = true ==> sets the columns of the matrix to be [v0 v1 v2]
	// otherwise, sets the rows
	Matrix3f( const Vector3f& v0, const Vector3f& v1, const Vector3f& v2, bool bCols = true );

	Matrix3f( const Matrix3f& rm ); // copy constructor
	Matrix3f& operator = ( const Matrix3f& rm ); // assignment operator
	// no destructor necessary

	// access the element at row i (mod 4), col j (mod 4)
	const float& operator () ( int i, int j ) const;
	float& operator () ( int i, int j );

	// get row i (mod 3)
	Vector3f getRow( int i ) const;
	void setRow( int i, const Vector3f& v );

	// get column j (mod 3)
	Vector3f getCol( int j ) const;
	void setCol( int j, const Vector3f& v );

	// gets the 2x2 submatrix of this matrix to m, with indices that wrap around
	// starting with upper left corner at (i0, j0)
	Matrix2f getSubmatrix2x2( int i0, int j0 ) const;

	// sets a 2x2 submatrix of this matrix to m, with indices that wrap around
	// starting with upper left corner at (i0, j0)
	void setSubmatrix2x2( int i0, int j0, const Matrix2f& m );

	float determinant() const;
	Matrix3f inverse( bool* pbIsSingular = NULL, float epsilon = 0.0001f ) const; // TODO: invert(), inverted()

	void transpose();
	Matrix3f transposed() const;

	// ---- Utility ----
	operator float* (); // automatic type conversion for GL
	void print();

	static float determinant3x3( float m00, float m01, float m02,
		float m10, float m11, float m12,
		float m20, float m21, float m22 );

	static Matrix3f ones();
	static Matrix3f identity();
	static Matrix3f rotation( const Vector3f& rDirection, float degrees );

	// Returns the rotation matrix represented by a unit quaternion
	// if q is not normalized, it it normalized first
	static Matrix3f rotation( const Quat4f& rq );

private:

	float m_afElements[9];

};

// Matrix-Vector multiplication
// 3x3 * 3x1 ==> 3x1
Vector3f operator * ( const Matrix3f& m, const Vector3f& v );

// Matrix-Matrix multiplication
Matrix3f operator * ( const Matrix3f& x, const Matrix3f& y );

#endif // MATRIX3F_H
