#include "GL/GLCamera.h"

#include <vecmath/Vector4f.h>

#include "geometry/BoundingBox3f.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

// static
Matrix4f GLCamera::makeProjectionMatrix( float fLeft, float fRight,
										float fBottom, float fTop,
										float fZNear, float fZFar )
{
	Matrix4f projection; // zero matrix

	projection( 0, 0 ) = ( 2.0f * fZNear ) / ( fRight - fLeft );
	projection( 1, 1 ) = ( 2.0f * fZNear ) / ( fTop - fBottom );
	projection( 0, 2 ) = ( fRight + fLeft ) / ( fRight - fLeft );
	projection( 1, 2 ) = ( fTop + fBottom ) / ( fTop - fBottom );

	projection( 2, 2 ) = -( fZFar + fZNear ) / ( fZFar - fZNear );
	projection( 2, 3 ) = -( 2.0f * fZFar * fZNear ) / ( fZFar - fZNear );

	projection( 3, 2 ) = -1;

	return projection;
}

// static
Matrix4f GLCamera::makeInfiniteProjectionMatrix( float fLeft, float fRight,
											 float fBottom, float fTop,
											 float fZNear )
{
	Matrix4f projection;

	projection( 0, 0 ) = ( 2.0f * fZNear ) / ( fRight - fLeft );
	projection( 1, 1 ) = ( 2.0f * fZNear ) / ( fTop - fBottom );
	projection( 0, 2 ) = ( fRight + fLeft ) / ( fRight - fLeft );
	projection( 1, 2 ) = ( fTop + fBottom ) / ( fTop - fBottom );

	// infinite view frustum
	// just take the limit as far --> inf of the regular frustum (else branch)
	projection( 2, 2 ) = -1.0f;
	projection( 2, 3 ) = -2.0f * fZNear;

	projection( 3, 2 ) = -1;

	return projection;
}

GLCamera::GLCamera( const Vector3f& vEye,
				   const Vector3f& vCenter,
				   const Vector3f& vUp,
				   float fLeft, float fRight,
				   float fBottom, float fTop,
				   float fZNear, float fZFar,
				   bool bIsInfinite )
{
	setLookAt( vEye, vCenter, vUp );
	setFrustum( fLeft, fRight,
		fBottom, fTop,
		fZNear, fZFar,
		false );
}

// virtual
GLCamera::~GLCamera()
{

}

void GLCamera::getFrustum( float* pfLeft, float* pfRight,
						  float* pfBottom, float* pfTop,
						  float* pfZNear, float* pfZFar,
						  bool* pbZFarIsInfinite ) const
{
	*pfLeft = m_fLeft;
	*pfRight = m_fRight;

	*pfBottom = m_fBottom;
	*pfTop = m_fTop;

	*pfZNear = m_fZNear;
	*pfZFar = m_fZFar;

	if( pbZFarIsInfinite != NULL )
	{
		*pbZFarIsInfinite = m_bZFarIsInfinite;
	}
}

bool GLCamera::isZFarInfinite()
{
	return m_bZFarIsInfinite;
}

void GLCamera::setFrustum( float left, float right,
				float bottom, float top,
				float zNear, float zFar,
				bool bZFarIsInfinite )
{
	m_fLeft = left;
	m_fRight = right;

	m_fBottom = bottom;
	m_fTop = top;

	m_fZNear = zNear;
	m_fZFar = zFar;

	m_bZFarIsInfinite = bZFarIsInfinite;
}

void GLCamera::getLookAt( Vector3f* pvEye,
						 Vector3f* pvCenter,
						 Vector3f* pvUp ) const
{
	*pvEye = m_vEye;
	*pvCenter = m_vCenter;
	*pvUp = m_vUp;
}

void GLCamera::setLookAt( const Vector3f& vEye,
						 const Vector3f& vCenter,
						 const Vector3f& vUp )
{
	m_vEye = vEye;
	m_vCenter = vCenter;
	m_vUp = vUp.normalized();
	
	// recompute up to ensure an orthonormal basis
	m_vUp = Vector3f::cross( -getForward(), getRight() );
}

void GLCamera::setEye( const Vector3f& vEye )
{
	setLookAt( vEye, m_vCenter, m_vUp );
}

void GLCamera::setCenter( const Vector3f& vCenter )
{
	setLookAt( m_vEye, vCenter, m_vUp );
}

void GLCamera::setUp( const Vector3f& vUp )
{
	setLookAt( m_vEye, m_vCenter, vUp );
}

Vector3f GLCamera::getUp() const
{
	return m_vUp;
}

Vector3f GLCamera::getRight() const
{
	return Vector3f::cross( getForward(), m_vUp );
}

Vector3f GLCamera::getForward() const
{
	return ( m_vCenter - m_vEye ).normalized();
}

float GLCamera::getZNear()
{
	return m_fZNear;
}

float GLCamera::getZFar()
{
	return m_fZFar;
}

Matrix4f GLCamera::getProjectionMatrix() const
{
	if( m_bZFarIsInfinite )
	{
		return makeInfiniteProjectionMatrix( m_fLeft, m_fRight,
			m_fBottom, m_fTop,
			m_fZNear );
	}
	else
	{
		return makeProjectionMatrix( m_fLeft, m_fRight,
			m_fBottom, m_fTop,
			m_fZNear, m_fZFar );
	}
}

Matrix4f GLCamera::getJitteredProjectionMatrix( float fEyeX, float fEyeY, float fFocusZ ) const
{
	float dx = -fEyeX * m_fZNear / fFocusZ;
	float dy = -fEyeY * m_fZNear / fFocusZ;

	if( m_bZFarIsInfinite )
	{
		return makeInfiniteProjectionMatrix( m_fLeft + dx, m_fRight + dx,
			m_fBottom + dy, m_fTop + dy,
			m_fZNear );
	}
	else
	{
		return makeProjectionMatrix( m_fLeft + dx, m_fRight + dx,
			m_fBottom + dy, m_fTop + dy,
			m_fZNear, m_fZFar );
	}
}

Matrix4f GLCamera::getViewMatrix() const
{
	Matrix4f view;

	// z is negative forward
	Vector3f z = -getForward();
	Vector3f y = getUp();
	Vector3f x = getRight();

	// the x, y, and z vectors define the orthonormal coordinate system
	// the affine part defines the overall translation

	view.setRow( 0, Vector4f( x, -Vector3f::dot( x, m_vEye ) ) );
	view.setRow( 1, Vector4f( y, -Vector3f::dot( y, m_vEye ) ) );
	view.setRow( 2, Vector4f( z, -Vector3f::dot( z, m_vEye ) ) );
	view.setRow( 3, Vector4f( 0, 0, 0, 1 ) );

	return view;
}

Matrix4f GLCamera::getJitteredViewMatrix( float fEyeX, float fEyeY ) const
{
	Matrix4f view;

	// z is negative forward
	Vector3f z = -getForward();
	Vector3f y = getUp();
	Vector3f x = getRight();

	// the x, y, and z vectors define the orthonormal coordinate system
	// the affine part defines the overall translation

	Vector3f jitteredEye = m_vEye + fEyeX * x + fEyeY * y;

	view.setRow( 0, Vector4f( x, -Vector3f::dot( x, jitteredEye ) ) );
	view.setRow( 1, Vector4f( y, -Vector3f::dot( y, jitteredEye ) ) );
	view.setRow( 2, Vector4f( z, -Vector3f::dot( z, jitteredEye ) ) );
	view.setRow( 3, Vector4f( 0, 0, 0, 1 ) );

	return view;
}

Matrix4f GLCamera::getViewProjectionMatrix() const
{
	return getProjectionMatrix() * getViewMatrix();
}

Matrix4f GLCamera::getJitteredViewProjectionMatrix( float fEyeX, float fEyeY, float fFocusZ ) const
{
	return
		(
			getJitteredProjectionMatrix( fEyeX, fEyeY, fFocusZ ) *
			getJitteredViewMatrix( fEyeX, fEyeY )
		);
}

Matrix4f GLCamera::getInverseProjectionMatrix() const
{
	return getProjectionMatrix().inverse();
}

Matrix4f GLCamera::getInverseViewMatrix() const
{
	return getViewMatrix().inverse();
}

Matrix4f GLCamera::getInverseViewProjectionMatrix() const
{
	return getViewProjectionMatrix().inverse();
}
