#include <cassert>
#include <cstdio>

#include <GL/glew.h>
#include <math/MathUtils.h>
#include <QImage>

#include "GL/GLTextureRectangle.h"
#include "GL/GLUtilities.h"
#include "io/PortableFloatMapIO.h"
#include "io/PortablePixelMapIO.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

// static
void GLTextureRectangle::setEnabled( bool bEnabled )
{
	if( bEnabled )
	{
		glEnable( GL_TEXTURE_RECTANGLE_ARB );
	}
	else
	{
		glDisable( GL_TEXTURE_RECTANGLE_ARB );
	}
}

// static
GLTextureRectangle* GLTextureRectangle::createDepthTexture( int width, int height, int nBits, const uint* auiData )
{
	GLTexture::GLTextureInternalFormat internalFormat;

	switch( nBits )
	{
	case 16:
		internalFormat = GLTexture::DEPTH_COMPONENT_FLOAT_16;
		break;
	case 24:
		internalFormat = GLTexture::DEPTH_COMPONENT_FLOAT_24;
		break;
	case 32:
		internalFormat = GLTexture::DEPTH_COMPONENT_FLOAT_32;
		break;
	default:
		fprintf( stderr, "Depth texture precision must be 16, 24, or 32 bits" );
		assert( false );
		break;
	}	

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, auiData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createUnsignedByte1Texture( int width, int height, const ubyte* aubData )
{
	GLTexture::GLTextureInternalFormat internalFormat = GLTexture::LUMINANCE_UNSIGNED_BYTE_8;

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height,
		internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, aubData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createUnsignedByte3Texture( int width, int height, const ubyte* aubData )
{
	printf( "GLTextureRectangle::unsigned byte 3: allocating %f megabytes\n", 3 * width * height / 1048576.f );

	GLTexture::GLTextureInternalFormat internalFormat = GLTexture::RGB_UNSIGNED_BYTE_8;

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, aubData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createUnsignedByte4Texture( int width, int height, const ubyte* aubData )
{
	GLTexture::GLTextureInternalFormat internalFormat = GLTexture::RGBA_UNSIGNED_BYTE_8;

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, aubData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createFloat1Texture( int width, int height, int nBitsPerComponent, const float* afData )
{
	printf( "GLTextureRectangle: allocating %f megabytes\n", 4 * width * height / 1048576.f );

	GLTexture::GLTextureInternalFormat internalFormat;

	switch( nBitsPerComponent )
	{
	case 16:
		internalFormat = GLTexture::LUMINANCE_FLOAT_16;
		// internalFormat = GLTexture::RED_FLOAT_16;
		break;
	case 32:
		internalFormat = GLTexture::LUMINANCE_FLOAT_32;
		// internalFormat = GLTexture::RED_FLOAT_32;
		break;
	default:
		fprintf( stderr, "Floating point texture nBits must be 16 or 32 bits!\n" );
		assert( false );
	}

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_LUMINANCE, GL_FLOAT, afData );
	// glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_RED, GL_FLOAT, afData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createFloat2Texture( int width, int height, int nBitsPerComponent, const float* afData )
{
	GLTexture::GLTextureInternalFormat internalFormat;

	switch( nBitsPerComponent )
	{
	case 16:
		internalFormat = GLTexture::LUMINANCE_ALPHA_FLOAT_16;
		break;
	case 32:
		internalFormat = GLTexture::LUMINANCE_ALPHA_FLOAT_32;
		break;
	default:
		fprintf( stderr, "Floating point texture nBits must be 16 or 32 bits!\n" );
		assert( false );
	}

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_LUMINANCE, GL_FLOAT, afData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createFloat3Texture( int width, int height, int nBits, const float* afData )
{
	GLTexture::GLTextureInternalFormat internalFormat;

	switch( nBits )
	{
	case 16:
		internalFormat = GLTexture::RGB_FLOAT_16;
		break;
	case 32:
		internalFormat = GLTexture::RGB_FLOAT_32;
		break;
	default:
		fprintf( stderr, "Floating point texture nBits must be 16 or 32 bits!\n" );
		assert( false );
	}

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_RGB, GL_FLOAT, afData );

	return pTexture;
}

// static
GLTextureRectangle* GLTextureRectangle::createFloat4Texture( int width, int height, int nBits, const float* afData )
{
	printf( "GLTextureRectangle: allocating %f megabytes\n", 16 * width * height / 1048576.f );

	GLTexture::GLTextureInternalFormat internalFormat;

	switch( nBits )
	{
	case 16:
		internalFormat = GLTexture::RGBA_FLOAT_16;
		break;
	case 32:
		internalFormat = GLTexture::RGBA_FLOAT_32;
		break;
	default:
		fprintf( stderr, "Floating point texture nBits must be 16 or 32 bits!\n" );
		assert( false );
	}

	GLTextureRectangle* pTexture = GLTextureRectangle::createTextureRectangle( width, height, internalFormat );

	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, internalFormat, width, height, 0, GL_RGBA, GL_FLOAT, afData );

	return pTexture;
}

void GLTextureRectangle::setFloat1Data( const float* afData )
{
	bind();
	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GLTexture::LUMINANCE_FLOAT_32, m_iWidth, m_iHeight, 0, GL_LUMINANCE, GL_FLOAT, afData );
}

void GLTextureRectangle::setFloat3Data( const float* afData )
{
	bind();
	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GLTexture::RGB_FLOAT_32, m_iWidth, m_iHeight, 0, GL_RGBA, GL_FLOAT, afData );
}

void GLTextureRectangle::setFloat4Data( const float* afData )
{
	bind();
	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GLTexture::RGBA_FLOAT_32, m_iWidth, m_iHeight, 0, GL_RGBA, GL_FLOAT, afData );
}

void GLTextureRectangle::setFloat1SubData( const float* afData, int x, int y, int width, int height )
{
	bind();
	glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, x, y, width, height, GL_LUMINANCE, GL_FLOAT, afData );
}

void GLTextureRectangle::setFloat3SubData( const float* afData, int x, int y, int width, int height )
{
	bind();
	glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, x, y, width, height, GL_RGB, GL_FLOAT, afData );
}

void GLTextureRectangle::setFloat4SubData( const float* afData, int x, int y, int width, int height )
{
	bind();
	glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, x, y, width, height, GL_RGBA, GL_FLOAT, afData );
}

void GLTextureRectangle::setUnsignedByte1Data( const GLubyte* aubData,
									   int xOffset, int yOffset,
									   int width, int height )
{
	bind();

	if( width == 0 )
	{
		width = m_iWidth;
	}
	if( height == 0 )
	{
		height = m_iHeight;
	}

	glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, xOffset, yOffset, width, height, GL_LUMINANCE, GL_UNSIGNED_BYTE, aubData	);
}

void GLTextureRectangle::setUnsignedByte3Data( const GLubyte* aubData,
									   int xOffset, int yOffset,
									   int width, int height )
{
	bind();

	if( width == 0 )
	{
		width = m_iWidth;
	}
	if( height == 0 )
	{
		height = m_iHeight;
	}

	// glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, xOffset, yOffset, width, height, GL_RGB, GL_UNSIGNED_BYTE, aubData	);
	glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB8, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, aubData );
}

void GLTextureRectangle::setUnsignedByte4Data( const GLubyte* aubData,
									   int xOffset, int yOffset,
									   int width, int height )
{
	bind();

	if( width == 0 )
	{
		width = m_iWidth;
	}
	if( height == 0 )
	{
		height = m_iHeight;
	}

	glTexSubImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, xOffset, yOffset, width, height, GL_RGBA, GL_UNSIGNED_BYTE, aubData );
}

int GLTextureRectangle::getWidth()
{
	return m_iWidth;
}

int GLTextureRectangle::getHeight()
{
	return m_iHeight;
}

void GLTextureRectangle::setAllWrapModes( GLint iMode )
{
	bind();
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, iMode );
	glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, iMode );
}

// virtual
void GLTextureRectangle::dumpToCSV( const char* szFilename )
{
	// TODO: use getData(), use it to get the right formats

	float* pixels = new float[ 4 * m_iWidth * m_iHeight ];
	getFloat4Data( pixels );

	int k = 0;
	FILE* fp = fopen( szFilename, "w" );
	fprintf( fp, "width = %d,height = %d\n", m_iWidth, m_iHeight );
	fprintf( fp, "pixel,byte,x,y,r,g,b,a\n" );

	for( int y = 0; y < m_iHeight; ++y )
	{
		for( int x = 0; x < m_iWidth; ++x )
		{
			fprintf( fp, "%d,%d,%d,%d,%f,%f,%f,%f\n", k / 4, k, x, y, pixels[k], pixels[k+1], pixels[k+2], pixels[k+3] );
			k += 4;
		}
	}

	fclose( fp );

	delete[] pixels;
}

// virtual
void GLTextureRectangle::dumpToTXT( const char* szFilename, GLint level, GLenum format, GLenum type )
{
	// TODO: use getData(), use it to get the right formats

	assert( level == 0 );
	assert( format == GL_RGBA );
	assert( type == GL_FLOAT );

	bind();

	float* pixels = new float[ 4 * m_iWidth * m_iHeight ];

	glGetTexImage( GL_TEXTURE_RECTANGLE_ARB, level, format, type, pixels );

	FILE* fp = fopen( szFilename, "w" );
	fprintf( fp, "width = %d, height = %d\n", m_iWidth, m_iHeight );

	for( int y = 0; y < m_iHeight; ++y )
	{
		for( int x = 0; x < m_iWidth; ++x )
		{
			int k = 4 * ( y * m_iWidth + x );

			fprintf( fp, "{%d} [%d] (%d, %d): <%f, %f, %f, %f>\n", k / 4, k, x, y, pixels[k], pixels[k+1], pixels[k+2], pixels[k+3] );
			k += 4;
		}
	}

	fclose( fp );

	delete[] pixels;
}

void GLTextureRectangle::dumpToTXTTiled( const char* szFilename, int gridWidth )
{
	// TODO: use getData(), use it to get the right formats
	bind();

	float* pixels = new float[ 4 * m_iWidth * m_iHeight ];

	glGetTexImage( GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA, GL_FLOAT, pixels );

	int k = 0;
	FILE* fp = fopen( szFilename, "w" );
	fprintf( fp, "width = %d, height = %d\n", m_iWidth, m_iHeight );

	for( int y = 0; y < m_iHeight; ++y )
	{
		for( int x = 0; x < m_iWidth; ++x )
		{
			int gx = x % gridWidth;
			int gz = x / gridWidth;

			fprintf( fp, "{%d} [%d] (%d, %d), ((%d, %d, %d)): <%f, %f, %f, %f>\n", k / 4, k, x, y,
				gx, y, gz,
				pixels[k], pixels[k+1], pixels[k+2], pixels[k+3] );
			k += 4;
		}
	}

	fclose( fp );

	delete[] pixels;
}

void GLTextureRectangle::dumpToPPM( const char* szFilename )
{
	bind();

	ubyte* pixels = new ubyte[ 3 * m_iWidth * m_iHeight ];
	getUnsignedByte3Data( pixels );
	PortablePixelMapIO::writeRGB( szFilename, pixels, m_iWidth, m_iHeight, true );
	delete[] pixels;
}

void GLTextureRectangle::dumpToPNG( const char* szFilename )
{
	ubyte* pixels = new ubyte[ 4 * m_iWidth * m_iHeight ];
	getUnsignedByte4Data( pixels );

	// TODO: arrayToQImage
	QImage q( m_iWidth, m_iHeight, QImage::Format_ARGB32 );
	for( int y = 0; y < m_iHeight; ++y )
	{
		for( int x = 0; x < m_iWidth; ++x )
		{
			// int yy = m_iHeight - y - 1;
			int yy = y;
			int k = 4 * ( yy * m_iWidth + x );

			ubyte r = pixels[ k ];
			ubyte g = pixels[ k + 1 ];
			ubyte b = pixels[ k + 2 ];

			q.setPixel( x, y, qRgb( r, g, b ) );
		}
	}

	delete[] pixels;
	q.save( szFilename, "PNG" );
}

void GLTextureRectangle::dumpToPFM( const char* szFilename )
{
	float* pixels = new float[ 3 * m_iWidth * m_iHeight ];
	getFloat3Data( pixels );

	PortableFloatMapIO::writeRGB( szFilename, pixels, m_iWidth, m_iHeight, -1.f,
		true );

	delete[] pixels;
}

//////////////////////////////////////////////////////////////////////////
// Private
//////////////////////////////////////////////////////////////////////////

// static
GLTextureRectangle* GLTextureRectangle::createTextureRectangle( int width, int height,
										  GLTexture::GLTextureInternalFormat internalFormat )
{
	GLTextureRectangle* pTexture = new GLTextureRectangle( width, height, internalFormat );
	pTexture->setAllWrapModes( GL_CLAMP );
	pTexture->setFilterMode( GLTexture::FILTER_MODE_NEAREST, GLTexture::FILTER_MODE_NEAREST );
	return pTexture;
}

GLTextureRectangle::GLTextureRectangle( int width, int height,
						 GLTexture::GLTextureInternalFormat internalFormat ) :

	m_iWidth( width ),
	m_iHeight( height ),

GLTexture( GL_TEXTURE_RECTANGLE_ARB, internalFormat )	
{
	assert( width > 0 );
	assert( height > 0 );
}
