#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>

#include "vecmath/Matrix2f.h"
#include "vecmath/Vector2f.h"

Matrix2f::Matrix2f()
{
	memset( m_afElements, 0, 4 * sizeof( float ) );
}

Matrix2f::Matrix2f( float m00, float m01,
				   float m10, float m11 )
{
	m_afElements[0] = m00;
	m_afElements[1] = m10;

	m_afElements[2] = m01;
	m_afElements[3] = m11;
}

Matrix2f::Matrix2f( const Vector2f& v0, const Vector2f& v1, bool bCols )
{
	if( bCols )
	{
		setCol( 0, v0 );
		setCol( 1, v1 );
	}
	else
	{
		setRow( 0, v0 );
		setRow( 1, v1 );
	}
}

Matrix2f::Matrix2f( const Matrix2f& rm )
{
	memcpy( m_afElements, rm.m_afElements, 2 * sizeof( float ) );
}

Matrix2f& Matrix2f::operator = ( const Matrix2f& rm )
{
	if( this != &rm )
	{
		memcpy( m_afElements, rm.m_afElements, 2 * sizeof( float ) );
	}
	return *this;
}

const float& Matrix2f::operator () ( int i, int j ) const
{
	i = i % 2;
	j = j % 2;
	return m_afElements[ j * 2 + i ];
}

float& Matrix2f::operator () ( int i, int j )
{
	i = i % 2;
	j = j % 2;
	return m_afElements[ j * 2 + i ];
}

Vector2f Matrix2f::getRow( int i ) const
{
	i = i % 2;

	return Vector2f
		(
			m_afElements[ i ],
			m_afElements[ i + 2 ]
		);
}

void Matrix2f::setRow( int i, const Vector2f& v )
{
	i = i % 2;

	m_afElements[ i ] = v.x();
	m_afElements[ i + 2 ] = v.y();
}

Vector2f Matrix2f::getCol( int j ) const
{
	int colStart = 2 * ( j % 2 );

	return Vector2f
		(
			m_afElements[ colStart ],
			m_afElements[ colStart + 1 ]
		);
}

void Matrix2f::setCol( int j, const Vector2f& v )
{
	int colStart = 2 * ( j % 2 );

	m_afElements[ colStart ] = v.x();
	m_afElements[ colStart + 1 ] = v.y();
}

float Matrix2f::determinant()
{
	return Matrix2f::determinant2x2
		(
			m_afElements[0], m_afElements[2],
			m_afElements[1], m_afElements[3]
		);
}

Matrix2f Matrix2f::inverse( bool* pbIsSingular, float epsilon )
{
	float determinant = m_afElements[0] * m_afElements[3] - m_afElements[2] * m_afElements[1];

	bool isSingular = ( fabs( determinant ) < epsilon );
	if( isSingular )
	{
		if( pbIsSingular != NULL )
		{
			*pbIsSingular = true;
		}
		return Matrix2f();
	}
	else
	{
		if( pbIsSingular != NULL )
		{
			*pbIsSingular = false;
		}

		float reciprocalDeterminant = 1.0f / determinant;

		return Matrix2f
			(
				m_afElements[3] * reciprocalDeterminant, -m_afElements[2] * reciprocalDeterminant,
				-m_afElements[1] * reciprocalDeterminant, m_afElements[0] * reciprocalDeterminant
			);
	}
}

// ---- Utility ----

Matrix2f::operator float* ()
{
	return m_afElements;
}

void Matrix2f::print()
{
	printf( "[ %.2f %.2f ]\n[ %.2f %.2f ]\n",
		m_afElements[0], m_afElements[2],
		m_afElements[1], m_afElements[3] );
}

// static
float Matrix2f::determinant2x2( float m00, float m01,
							   float m10, float m11 )
{
	return( m00 * m11 - m01 * m10 );
}

// static
Matrix2f Matrix2f::ones()
{
	Matrix2f m;
	for( int i = 0; i < 4; ++i )
	{
		m.m_afElements[i] = 1;
	}

	return m;
}

// static
Matrix2f Matrix2f::identity()
{
	Matrix2f m;

	m( 0, 0 ) = 1;
	m( 1, 1 ) = 1;

	return m;
}

// static
Matrix2f Matrix2f::rotation( float degrees )
{
	float c = cos( degrees );
	float s = sin( degrees );

	return Matrix2f
		(
			c, -s,
			s, c
		);
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

Vector2f operator * ( const Matrix2f& m, const Vector2f& v )
{
	Vector2f output( 0, 0 );

	for( int i = 0; i < 2; ++i )
	{
		for( int j = 0; j < 2; ++j )
		{
			output[i] += m( i, j ) * v[ j ];
		}
	}

	return output;
}

Matrix2f operator * ( const Matrix2f& x, const Matrix2f& y )
{
	Matrix2f product; // zeroes

	for( int i = 0; i < 2; ++i )
	{
		for( int j = 0; j < 2; ++j )
		{
			for( int k = 0; k < 2; ++k )
			{
				product( i, k ) += x( i, j ) * y( j, k );
			}
		}
	}

	return product;
}
