#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstring>

#include <math/MathUtils.h>

#include "vecmath/Matrix4f.h"

#include "vecmath/Matrix2f.h"
#include "vecmath/Matrix3f.h"
#include "vecmath/Quat4f.h"
#include "vecmath/Vector3f.h"
#include "vecmath/Vector4f.h"

Matrix4f::Matrix4f()
{
	memset( m_afElements, 0, 16 * sizeof( float ) );
}

Matrix4f::Matrix4f( float m00, float m01, float m02, float m03,
				   float m10, float m11, float m12, float m13,
				   float m20, float m21, float m22, float m23,
				   float m30, float m31, float m32, float m33 )
{
	m_afElements[0] = m00;
	m_afElements[1] = m10;
	m_afElements[2] = m20;
	m_afElements[3] = m30;
	
	m_afElements[4] = m01;
	m_afElements[5] = m11;
	m_afElements[6] = m21;
	m_afElements[7] = m31;

	m_afElements[8] = m02;
	m_afElements[9] = m12;
	m_afElements[10] = m22;
	m_afElements[11] = m32;

	m_afElements[12] = m03;
	m_afElements[13] = m13;
	m_afElements[14] = m23;
	m_afElements[15] = m33;
}

Matrix4f::Matrix4f( const Vector4f& v0, const Vector4f& v1, const Vector4f& v2, const Vector4f& v3, bool bCols )
{
	if( bCols )
	{
		setCol( 0, v0 );
		setCol( 1, v1 );
		setCol( 2, v2 );
		setCol( 3, v3 );
	}
	else
	{
		setRow( 0, v0 );
		setRow( 1, v1 );
		setRow( 2, v2 );
		setRow( 3, v3 );
	}
}

Matrix4f::Matrix4f( const Matrix4f& rm )
{
	memcpy( m_afElements, rm.m_afElements, 16 * sizeof( float ) );
}

Matrix4f& Matrix4f::operator = ( const Matrix4f& rm )
{
	if( this != &rm )
	{
		memcpy( m_afElements, rm.m_afElements, 16 * sizeof( float ) );
	}
	return *this;
}

const float& Matrix4f::operator () ( int i, int j ) const
{
	i = i % 4;
	j = j % 4;
	return m_afElements[ j * 4 + i ];
}

float& Matrix4f::operator () ( int i, int j )
{
	i = i % 4;
	j = j % 4;
	return m_afElements[ j * 4 + i ];
}

Vector4f Matrix4f::getRow( int i ) const
{
	i = i % 4;

	return Vector4f
		(
			m_afElements[ i ],
			m_afElements[ i + 4 ],
			m_afElements[ i + 8 ],
			m_afElements[ i + 12 ]
		);
}

void Matrix4f::setRow( int i, const Vector4f& v )
{
	i = i % 4;

	m_afElements[ i ] = v.x();
	m_afElements[ i + 4 ] = v.y();
	m_afElements[ i + 8 ] = v.z();
	m_afElements[ i + 12 ] = v.w();
}

Vector4f Matrix4f::getCol( int j ) const
{
	int colStart = 4 * ( j % 4 );

	return Vector4f
		(
			m_afElements[ colStart ],
			m_afElements[ colStart + 1 ],
			m_afElements[ colStart + 2 ],
			m_afElements[ colStart + 3 ]
		);
}

void Matrix4f::setCol( int j, const Vector4f& v )
{
	int colStart = 4 * ( j % 4 );

	m_afElements[ colStart ] = v.x();
	m_afElements[ colStart + 1 ] = v.y();
	m_afElements[ colStart + 2 ] = v.z();
	m_afElements[ colStart + 3 ] = v.w();
}

Matrix2f Matrix4f::getSubmatrix2x2( int i0, int j0 ) const
{
	Matrix2f out;

	for( int i = 0; i < 2; ++i )
	{
		for( int j = 0; j < 2; ++j )
		{
			out( i, j ) = ( *this )( i + i0, j + j0 );
		}
	}

	return out;
}

Matrix3f Matrix4f::getSubmatrix3x3( int i0, int j0 ) const
{
	Matrix3f out;

	for( int i = 0; i < 3; ++i )
	{
		for( int j = 0; j < 3; ++j )
		{
			out( i, j ) = ( *this )( i + i0, j + j0 );
		}
	}

	return out;
}

void Matrix4f::setSubmatrix2x2( int i0, int j0, const Matrix2f& m )
{
	for( int i = 0; i < 2; ++i )
	{
		for( int j = 0; j < 2; ++j )
		{
			( *this )( i + i0, j + j0 ) = m( i, j );
		}
	}
}

void Matrix4f::setSubmatrix3x3( int i0, int j0, const Matrix3f& m )
{
	for( int i = 0; i < 3; ++i )
	{
		for( int j = 0; j < 3; ++j )
		{
			( *this )( i + i0, j + j0 ) = m( i, j );
		}
	}
}

float Matrix4f::determinant() const
{
	float m00 = m_afElements[0];
	float m10 = m_afElements[1];
	float m20 = m_afElements[2];
	float m30 = m_afElements[3];

	float m01 = m_afElements[4];
	float m11 = m_afElements[5];
	float m21 = m_afElements[6];
	float m31 = m_afElements[7];

	float m02 = m_afElements[8];
	float m12 = m_afElements[9];
	float m22 = m_afElements[10];
	float m32 = m_afElements[11];

	float m03 = m_afElements[12];
	float m13 = m_afElements[13];
	float m23 = m_afElements[14];
	float m33 = m_afElements[15];

	float cofactor00 =  Matrix3f::determinant3x3( m11, m12, m13, m21, m22, m23, m31, m32, m33 );
	float cofactor01 = -Matrix3f::determinant3x3( m12, m13, m10, m22, m23, m20, m32, m33, m30 );
	float cofactor02 =  Matrix3f::determinant3x3( m13, m10, m11, m23, m20, m21, m33, m30, m31 );
	float cofactor03 = -Matrix3f::determinant3x3( m10, m11, m12, m20, m21, m22, m30, m31, m32 );

	return( m00 * cofactor00 + m01 * cofactor01 + m02 * cofactor02 + m03 * cofactor03 );
}

Matrix4f Matrix4f::inverse( bool* pbIsSingular, float epsilon ) const
{
	float m00 = m_afElements[0];
	float m10 = m_afElements[1];
	float m20 = m_afElements[2];
	float m30 = m_afElements[3];

	float m01 = m_afElements[4];
	float m11 = m_afElements[5];
	float m21 = m_afElements[6];
	float m31 = m_afElements[7];

	float m02 = m_afElements[8];
	float m12 = m_afElements[9];
	float m22 = m_afElements[10];
	float m32 = m_afElements[11];

	float m03 = m_afElements[12];
	float m13 = m_afElements[13];
	float m23 = m_afElements[14];
	float m33 = m_afElements[15];

    float cofactor00 =  Matrix3f::determinant3x3( m11, m12, m13, m21, m22, m23, m31, m32, m33 );
    float cofactor01 = -Matrix3f::determinant3x3( m12, m13, m10, m22, m23, m20, m32, m33, m30 );
    float cofactor02 =  Matrix3f::determinant3x3( m13, m10, m11, m23, m20, m21, m33, m30, m31 );
    float cofactor03 = -Matrix3f::determinant3x3( m10, m11, m12, m20, m21, m22, m30, m31, m32 );
    
    float cofactor10 = -Matrix3f::determinant3x3( m21, m22, m23, m31, m32, m33, m01, m02, m03 );
    float cofactor11 =  Matrix3f::determinant3x3( m22, m23, m20, m32, m33, m30, m02, m03, m00 );
    float cofactor12 = -Matrix3f::determinant3x3( m23, m20, m21, m33, m30, m31, m03, m00, m01 );
    float cofactor13 =  Matrix3f::determinant3x3( m20, m21, m22, m30, m31, m32, m00, m01, m02 );
    
    float cofactor20 =  Matrix3f::determinant3x3( m31, m32, m33, m01, m02, m03, m11, m12, m13 );
    float cofactor21 = -Matrix3f::determinant3x3( m32, m33, m30, m02, m03, m00, m12, m13, m10 );
    float cofactor22 =  Matrix3f::determinant3x3( m33, m30, m31, m03, m00, m01, m13, m10, m11 );
    float cofactor23 = -Matrix3f::determinant3x3( m30, m31, m32, m00, m01, m02, m10, m11, m12 );
    
    float cofactor30 = -Matrix3f::determinant3x3( m01, m02, m03, m11, m12, m13, m21, m22, m23 );
    float cofactor31 =  Matrix3f::determinant3x3( m02, m03, m00, m12, m13, m10, m22, m23, m20 );
    float cofactor32 = -Matrix3f::determinant3x3( m03, m00, m01, m13, m10, m11, m23, m20, m21 );
    float cofactor33 =  Matrix3f::determinant3x3( m00, m01, m02, m10, m11, m12, m20, m21, m22 );

	float determinant = m00 * cofactor00 + m01 * cofactor01 + m02 * cofactor02 + m03 * cofactor03;

	bool isSingular = ( fabs( determinant ) < epsilon );
	if( isSingular )
	{
		if( pbIsSingular != NULL )
		{
			*pbIsSingular = true;
		}
		return Matrix4f();
	}
	else
	{
		if( pbIsSingular != NULL )
		{
			*pbIsSingular = false;
		}

		float reciprocalDeterminant = 1.0f / determinant;

		return Matrix4f
			(
				cofactor00 * reciprocalDeterminant, cofactor10 * reciprocalDeterminant, cofactor20 * reciprocalDeterminant, cofactor30 * reciprocalDeterminant,
				cofactor01 * reciprocalDeterminant, cofactor11 * reciprocalDeterminant, cofactor21 * reciprocalDeterminant, cofactor31 * reciprocalDeterminant,
				cofactor02 * reciprocalDeterminant, cofactor12 * reciprocalDeterminant, cofactor22 * reciprocalDeterminant, cofactor32 * reciprocalDeterminant,
				cofactor03 * reciprocalDeterminant, cofactor13 * reciprocalDeterminant, cofactor23 * reciprocalDeterminant, cofactor33 * reciprocalDeterminant
			);
	}
}

void Matrix4f::transpose()
{
	float temp;

	for( int i = 0; i < 3; ++i )
	{
		for( int j = i + 1; j < 4; ++j )
		{
			temp = ( *this )( i, j );
			( * this )( i, j ) = ( *this )( j, i );
			( *this )( j, i ) = temp;
		}
	}
}

Matrix4f Matrix4f::transposed() const
{
	Matrix4f out;
	for( int i = 0; i < 4; ++i )
	{
		for( int j = 0; j < 4; ++j )
		{
			out( j, i ) = ( *this )( i, j );
		}
	}

	return out;
}

// ---- Utility ----

Matrix4f::operator float* ()
{
	return m_afElements;
}

void Matrix4f::print()
{
	printf( "[ %.2f %.2f %.2f %.2f ]\n[ %.2f %.2f %.2f %.2f ]\n[ %.2f %.2f %.2f %.2f ]\n[ %.2f %.2f %.2f %.2f ]\n",
		m_afElements[0], m_afElements[4], m_afElements[8], m_afElements[12],
		m_afElements[1], m_afElements[5], m_afElements[9], m_afElements[13],
		m_afElements[2], m_afElements[6], m_afElements[10], m_afElements[14],
		m_afElements[3], m_afElements[7], m_afElements[11], m_afElements[15] );
}

// static
Matrix4f Matrix4f::ones()
{
	Matrix4f m;
	for( int i = 0; i < 16; ++i )
	{
		m.m_afElements[i] = 1;
	}

	return m;
}

// static
Matrix4f Matrix4f::identity()
{
	Matrix4f m;
	
	m( 0, 0 ) = 1;
	m( 1, 1 ) = 1;
	m( 2, 2 ) = 1;
	m( 3, 3 ) = 1;

	return m;
}

// static
Matrix4f Matrix4f::translation( const Vector3f& rTranslation )
{
	return Matrix4f
		(
			1, 0, 0, rTranslation.x(),
			0, 1, 0, rTranslation.y(),
			0, 0, 1, rTranslation.z(),
			0, 0, 0, 1
		);
}

// static
Matrix4f Matrix4f::rotation( const Vector3f& rDirection, float degrees )
{
	Vector3f normalizedDirection = rDirection.normalized();
	
	float theta = MathUtils::degreesToRadians( degrees );
	float cosTheta = cos( theta );
	float sinTheta = sin( theta );

	float x = normalizedDirection.x();
	float y = normalizedDirection.y();
	float z = normalizedDirection.z();

	return Matrix4f
		(
			x * x * ( 1.0f - cosTheta ) + cosTheta,			y * x * ( 1.0f - cosTheta ) - z * sinTheta,		z * x * ( 1.0f - cosTheta ) + y * sinTheta,		0.0f,
			x * y * ( 1.0f - cosTheta ) + z * sinTheta,		y * y * ( 1.0f - cosTheta ) + cosTheta,			z * y * ( 1.0f - cosTheta ) - x * sinTheta,		0.0f,
			x * z * ( 1.0f - cosTheta ) - y * sinTheta,		y * z * ( 1.0f - cosTheta ) + x * sinTheta,		z * z * ( 1.0f - cosTheta ) + cosTheta,			0.0f,
			0.0f,											0.0f,											0.0f,											1.0f
		);
}

// static
Matrix4f Matrix4f::rotation( const Quat4f& rq )
{
	Quat4f q = rq.normalized();

	float xx = q.x() * q.x();
	float yy = q.y() * q.y();
	float zz = q.z() * q.z();

	float xy = q.x() * q.y();
	float zw = q.z() * q.w();

	float xz = q.x() * q.z();
	float yw = q.y() * q.w();

	float yz = q.y() * q.z();
	float xw = q.x() * q.w();

	return Matrix4f
		(
			1.0f - 2.0f * ( yy + zz ),		2.0f * ( xy - zw ),				2.0f * ( xz + yw ),				0.0f,
			2.0f * ( xy + zw ),				1.0f - 2.0f * ( xx + zz ),		2.0f * ( yz - xw ),				0.0f,
			2.0f * ( xz - yw ),				2.0f * ( yz + xw ),				1.0f - 2.0f * ( xx + yy ),		0.0f,
			0.0f,							0.0f,							0.0f,							1.0f
		);
}

// static
Matrix4f Matrix4f::randomRotation( float u0, float u1, float u2 )
{
	return Matrix4f::rotation( Quat4f::randomRotation( u0, u1, u2 ) );
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

Vector4f operator * ( const Matrix4f& m, const Vector4f& v )
{
	Vector4f output( 0, 0, 0, 0 );

	for( int i = 0; i < 4; ++i )
	{
		for( int j = 0; j < 4; ++j )
		{
			output[i] += m( i, j ) * v[ j ];
		}
	}

	return output;
}

Matrix4f operator * ( const Matrix4f& x, const Matrix4f& y )
{
	Matrix4f product; // zeroes

	for( int i = 0; i < 4; ++i )
	{
		for( int j = 0; j < 4; ++j )
		{
			for( int k = 0; k < 4; ++k )
			{
				product( i, k ) += x( i, j ) * y( j, k );
			}
		}
	}

	return product;
}
