#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>

#include "vecmath/Vector2i.h"
#include "vecmath/Vector2f.h"
#include "vecmath/Vector3i.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

Vector2i::Vector2i()
{
	m_aiElements[0] = 0;
	m_aiElements[1] = 0;
}

Vector2i::Vector2i( int x, int y )
{
	m_aiElements[0] = x;
	m_aiElements[1] = y;
}

Vector2i::Vector2i( const Vector2i& rv )
{
	m_aiElements[0] = rv.m_aiElements[0];
	m_aiElements[1] = rv.m_aiElements[1];
}

Vector2i& Vector2i::operator = ( const Vector2i& rv )
{
	if( this != &rv )
	{
		m_aiElements[0] = rv.m_aiElements[0];
		m_aiElements[1] = rv.m_aiElements[1];
	}
	return *this;
}

const int& Vector2i::operator [] ( int i ) const
{
	return m_aiElements[ i % 2 ];
}

int& Vector2i::operator [] ( int i )
{
	return m_aiElements[ i % 2 ];
}

int Vector2i::x() const
{
	return m_aiElements[0];
}

int Vector2i::y() const
{
	return m_aiElements[1];
}

Vector2i Vector2i::xy() const
{
	return Vector2i( m_aiElements[0], m_aiElements[1] );
}

Vector2i Vector2i::yx() const
{
	return Vector2i( m_aiElements[1], m_aiElements[0] );
}

Vector2i Vector2i::xx() const
{
	return Vector2i( m_aiElements[0], m_aiElements[0] );
}

Vector2i Vector2i::yy() const
{
	return Vector2i( m_aiElements[1], m_aiElements[1] );
}

float Vector2i::abs() const
{
	return sqrt( static_cast< float >( absSquared() ) );
}

int Vector2i::absSquared() const
{
	return( m_aiElements[0] * m_aiElements[0] + m_aiElements[1] * m_aiElements[1] );
}

Vector2f Vector2i::normalized() const
{
	float length = abs();
	return Vector2f( m_aiElements[0] / length, m_aiElements[1] / length );
}

void Vector2i::negate()
{
	m_aiElements[0] = -m_aiElements[0];
	m_aiElements[1] = -m_aiElements[1];
}

// ---- Utility ----
Vector2i::operator const int* ()
{
	return m_aiElements;
}

void Vector2i::print() const
{
	printf( "< %1.2i, %1.2i >\n",
		m_aiElements[0], m_aiElements[1] );
}

// static
int Vector2i::dot( const Vector2i& v0, const Vector2i& v1 )
{
	return v0.x() * v1.x() + v0.y() * v1.y();
}

//static
Vector3i Vector2i::cross( const Vector2i& v0, const Vector2i& v1 )
{
	return Vector3i
	(
		0,
		0,
		v0.x() * v1.y() - v0.y() * v1.x()
	);
}

// static
Vector2f Vector2i::lerp( const Vector2i& v0, const Vector2i& v1, float alpha )
{
	return alpha * ( v1 - v0 ) + Vector2f( v0 );
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

bool operator == ( const Vector2i& v0, const Vector2i& v1 )
{
	return
	(
		v0.x() == v1.x() &&
		v0.y() == v1.y()
	);
}

Vector2i operator + ( const Vector2i& v0, const Vector2i& v1 )
{
	return Vector2i( v0.x() + v1.x(), v0.y() + v1.y() );
}

Vector2i operator - ( const Vector2i& v0, const Vector2i& v1 )
{
	return Vector2i( v0.x() - v1.x(), v0.y() - v1.y() );
}

Vector2i operator * ( const Vector2i& v0, const Vector2i& v1 )
{
	return Vector2i( v0.x() * v1.x(), v0.y() * v1.y() );
}

Vector2i operator / ( const Vector2i& v0, const Vector2i& v1 )
{
	return Vector2i( v0.x() / v1.x(), v0.y() / v1.y() );
}

Vector2i operator - ( const Vector2i& v )
{
	return Vector2i( -v.x(), -v.y() );
}

Vector2i operator * ( int c, const Vector2i& v )
{
	return Vector2i( c * v.x(), c * v.y() );
}

Vector2i operator * ( const Vector2i& v, int c )
{
	return Vector2i( c * v.x(), c * v.y() );
}

Vector2f operator * ( float f, const Vector2i& v )
{
	return Vector2f( f * v.x(), f * v.y() );
}

Vector2f operator * ( const Vector2i& v, float f )
{
	return Vector2f( f * v.x(), f * v.y() );
}
