#include <cmath>
#include <cstdio>
#include <cstdlib>

#include "vecmath/Vector3f.h"
#include "vecmath/Vector3d.h"
#include "vecmath/Vector3i.h"
#include "vecmath/Vector2f.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

Vector3f::Vector3f()
{
	m_afElements[0] = 0;
	m_afElements[1] = 0;
	m_afElements[2] = 0;
}

Vector3f::Vector3f( float x, float y, float z )
{
	m_afElements[0] = x;
	m_afElements[1] = y;
	m_afElements[2] = z;
}

Vector3f::Vector3f( const Vector2f& xy, float z )
{
	m_afElements[0] = xy.x();
	m_afElements[1] = xy.y();
	m_afElements[2] = z;
}

Vector3f::Vector3f( float x, const Vector2f& yz )
{
	m_afElements[0] = x;
	m_afElements[1] = yz.x();
	m_afElements[2] = yz.y();
}

Vector3f::Vector3f( const Vector3f& rv )
{
	m_afElements[0] = rv.m_afElements[0];
	m_afElements[1] = rv.m_afElements[1];
	m_afElements[2] = rv.m_afElements[2];
}

Vector3f::Vector3f( const Vector3d& rv )
{
	m_afElements[0] = rv.x();
	m_afElements[1] = rv.y();
	m_afElements[2] = rv.z();
}

Vector3f::Vector3f( const Vector3i& rv )
{
	m_afElements[0] = rv.x();
	m_afElements[1] = rv.y();
	m_afElements[2] = rv.z();
}

Vector3f& Vector3f::operator = ( const Vector3f& rv )
{
	if( this != &rv )
	{
		m_afElements[0] = rv.m_afElements[0];
		m_afElements[1] = rv.m_afElements[1];
		m_afElements[2] = rv.m_afElements[2];
	}
	return *this;
}

Vector3f& Vector3f::operator = ( const Vector3d& rv )
{
	m_afElements[ 0 ] = static_cast< float >( rv.x() );
	m_afElements[ 1 ] = static_cast< float >( rv.y() );
	m_afElements[ 2 ] = static_cast< float >( rv.z() );

	return *this;
}

Vector3f& Vector3f::operator = ( const Vector3i& rv )
{
	m_afElements[ 0 ] = rv.x();
	m_afElements[ 1 ] = rv.y();
	m_afElements[ 2 ] = rv.z();

	return *this;
}

const float& Vector3f::operator [] ( int i ) const
{
	return m_afElements[ i % 3 ];
}

float& Vector3f::operator [] ( int i )
{
	return m_afElements[ i % 3 ];
}

float Vector3f::x() const
{
	return m_afElements[0];
}

float Vector3f::y() const
{
	return m_afElements[1];
}

float Vector3f::z() const
{
	return m_afElements[2];
}

Vector2f Vector3f::xy() const
{
	return Vector2f( m_afElements[0], m_afElements[1] );
}

Vector2f Vector3f::xz() const
{
	return Vector2f( m_afElements[0], m_afElements[2] );
}

Vector2f Vector3f::yz() const
{
	return Vector2f( m_afElements[1], m_afElements[2] );
}

Vector3f Vector3f::xyz() const
{
	return Vector3f( m_afElements[0], m_afElements[1], m_afElements[2] );
}

Vector3f Vector3f::yzx() const
{
	return Vector3f( m_afElements[1], m_afElements[2], m_afElements[0] );
}

Vector3f Vector3f::zxy() const
{
	return Vector3f( m_afElements[2], m_afElements[0], m_afElements[1] );
}

float Vector3f::abs() const
{
	return sqrt( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] + m_afElements[2] * m_afElements[2] );
}

float Vector3f::absSquared() const
{
	return( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] + m_afElements[2] * m_afElements[2] );
}

void Vector3f::normalize()
{
	float norm = sqrt( m_afElements[0] * m_afElements[0] + m_afElements[1] * m_afElements[1] + m_afElements[2] * m_afElements[2] );
	m_afElements[0] = m_afElements[0] / norm;
	m_afElements[1] = m_afElements[1] / norm;
	m_afElements[2] = m_afElements[2] / norm;
}

Vector3f Vector3f::normalized() const
{
	float length = abs();
	return Vector3f
		(
			m_afElements[0] / length,
			m_afElements[1] / length,
			m_afElements[2] / length
		);
}

void Vector3f::negate()
{
	m_afElements[0] = -m_afElements[0];
	m_afElements[1] = -m_afElements[1];
	m_afElements[2] = -m_afElements[2];
}

// ---- Utility ----

Vector3f::operator const float* ()
{
	return m_afElements;
}

void Vector3f::print() const
{
	printf( "< %1.2f, %1.2f, %1.2f >\n",
		m_afElements[0], m_afElements[1], m_afElements[2] );
}

// static
float Vector3f::dot( const Vector3f& v0, const Vector3f& v1 )
{
	return v0.x() * v1.x() + v0.y() * v1.y() + v0.z() * v1.z();
}

// static
Vector3f Vector3f::cross( const Vector3f& v0, const Vector3f& v1 )
{
	return Vector3f
		(
			v0.y() * v1.z() - v0.z() * v1.y(),
			v0.z() * v1.x() - v0.x() * v1.z(),
			v0.x() * v1.y() - v0.y() * v1.x()
		);
}

// static
Vector3f Vector3f::lerp( const Vector3f& v0, const Vector3f& v1, float alpha )
{
	return alpha * ( v1 - v0 ) + v0;
}

//////////////////////////////////////////////////////////////////////////
// Operators
//////////////////////////////////////////////////////////////////////////

Vector3f operator + ( const Vector3f& v0, const Vector3f& v1 )
{
	return Vector3f( v0.x() + v1.x(), v0.y() + v1.y(), v0.z() + v1.z() );
}

Vector3f operator - ( const Vector3f& v0, const Vector3f& v1 )
{
	return Vector3f( v0.x() - v1.x(), v0.y() - v1.y(), v0.z() - v1.z() );
}

Vector3f operator * ( const Vector3f& v0, const Vector3f& v1 )
{
	return Vector3f( v0.x() * v1.x(), v0.y() * v1.y(), v0.z() * v1.z() );
}

Vector3f operator / ( const Vector3f& v0, const Vector3f& v1 )
{
	return Vector3f( v0.x() / v1.x(), v0.y() / v1.y(), v0.z() / v1.z() );
}

Vector3f operator - ( const Vector3f& v )
{
	return Vector3f( -v.x(), -v.y(), -v.z() );
}

Vector3f operator * ( float f, const Vector3f& v )
{
	return Vector3f( f * v.x(), f * v.y(), f * v.z() );
}

Vector3f operator * ( const Vector3f& v, float f )
{
	return Vector3f( f * v.x(), f * v.y(), f * v.z() );
}
