#include "video/FFMPEGVideo.h"

#include <cassert>

#include "math/Arithmetic.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

// static
Reference< IVideo > FFMPEGVideo::fromFile( QString filename )
{
	// one time registration
	if( !( FFMPEGVideo::s_bInitialized ) )
	{
		av_register_all();
		FFMPEGVideo::s_bInitialized = true;
	}

	int retVal;

	AVFormatContext* pFormatContext;
    AVCodecContext* pCodecContext;
    AVCodec* pCodec;
	AVFrame* pFrameRaw;
	AVFrame* pFrameRGB;	

	// Open the file and examine the header
	// populates pFormatContext
	retVal = av_open_input_file
		(
			&pFormatContext, // output context
			qPrintable( filename ), // filename
			NULL, // format, NULL --> auto-detect, otherwise, forces file format
			0, // buffer size, 0 --> auto-select
			NULL // format options, NULL --> auto-detect, otherwise force decode options
		);
	if( retVal == 0 ) // if succeeded
	{
		// Retrieve stream information
		// populates pFormatContext->streams with data

		retVal = av_find_stream_info( pFormatContext );
		if( retVal >= 0 ) // if succeeded
		{
			// TODO: let the user select which stream
			// Find the first video stream
			uint i = 0;
			int videoStreamIndex = -1;
			while( ( videoStreamIndex == -1 ) && ( i < pFormatContext->nb_streams ) )
			{
				if( pFormatContext->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO )
				{
					videoStreamIndex = i;
				}
				++i;
			}

			// if we found a video stream
			// load its codec context
			if( videoStreamIndex > -1 )
			{
				// get a pointer to the codec context for the video stream
				pCodecContext = pFormatContext->streams[ videoStreamIndex ]->codec;

				// find a codec for the codec context
				pCodec = avcodec_find_decoder( pCodecContext->codec_id );
				if( pCodec != NULL )
				{
					// ok we found a codec, try opening it
					retVal = avcodec_open( pCodecContext, pCodec );
					if( retVal >= 0 )
					{
						// Allocate a frame for the incoming data						
						pFrameRaw = avcodec_alloc_frame();
						if( pFrameRaw != NULL )
						{
							// Allocate another for
							pFrameRGB = avcodec_alloc_frame();
							if( pFrameRGB != NULL )
							{								
								// Note: PixelFormats are in avutil.h
								// Note: flags are in swscale.h
								SwsContext* pSWSContext = sws_getContext
									(
										pCodecContext->width, pCodecContext->height, // source width and height
										pCodecContext->pix_fmt, // source format
										pCodecContext->width, pCodecContext->height, // destination width and height
										PIX_FMT_RGB24, // destination format
										SWS_POINT, // flags
										NULL, // source filter, NULL --> default
										NULL, // destination filter, NULL --> default
										NULL // filter parameters, NULL --> default
									);

								if( pSWSContext != NULL )
								{
									return new FFMPEGVideo
									(
										pFormatContext,
										videoStreamIndex,
										pCodecContext,
										pFrameRaw,
										pFrameRGB,
										pSWSContext
									);
								}
								else
								{
									fprintf( stderr, "Error creating RGB conversion context!\n" );
								}

								av_free( pFrameRGB );
							}
							else
							{
								fprintf( stderr, "Error allocating RGB frame!\n" );
							}
							
							av_free( pFrameRaw );
						}
					}
					else
					{
						fprintf( stderr, "Error opening codec!\n" );
					}
				}
				else
				{
					fprintf( stderr, "Unsupported codec!\n" );
				}
			}			
			else
			{
				fprintf( stderr, "File contains no video streams!\n" );
			}
		}
		else
		{
			fprintf( stderr, "Error parsing stream information!\n" );
		}

		// close the video file in case of failure
		av_close_input_file( pFormatContext );
	}
	else
	{
		fprintf( stderr, "Error opening %s!\n", qPrintable( filename ) );
	}
	
	assert( false );
	return NULL;	
}

// virtual
FFMPEGVideo::~FFMPEGVideo()
{
	sws_freeContext( m_pSWSContext );
	av_free( m_pFrameRGB );
	av_free( m_pFrameRaw );
	avcodec_close( m_pCodecContext );
	av_close_input_file( m_pFormatContext );
}

// virtual
int64 FFMPEGVideo::numFrames()
{
	return m_pFormatContext->streams[ m_iVideoStreamIndex ]->duration;	
}

// virtual
int FFMPEGVideo::framePeriodMilliseconds()
{
	float seconds = framePeriodSeconds();
	return Arithmetic::roundToInt( seconds * 1000.f );
}

// virtual
float FFMPEGVideo::framePeriodSeconds()
{
	AVRational framePeriod = m_pFormatContext->streams[ m_iVideoStreamIndex ]->time_base;
	float seconds = static_cast< float >( av_q2d( framePeriod ) );
	return seconds;
}

// virtual
int FFMPEGVideo::width()
{
	return m_iWidth;
}

// virtual
int FFMPEGVideo::height()
{
	return m_iHeight;
}

// virtual
int FFMPEGVideo::bytesPerFrame()
{
	return( avpicture_get_size( PIX_FMT_RGB24, m_pCodecContext->width, m_pCodecContext->height ) );
}

// virtual
int64 FFMPEGVideo::getNextFrameIndex()
{
	return m_iNextFrameIndex;
}

// virtual
void FFMPEGVideo::setNextFrameIndex( int64 frameIndex )
{
	assert( ( frameIndex >= 0 ) &&
		( frameIndex < numFrames() ) );

	if( frameIndex != m_iNextFrameIndex + 1 )
	{
		m_bSeeking = true;
	}
	m_iNextFrameIndex = frameIndex;	

}

// virtual
bool FFMPEGVideo::getNextFrame( UnsignedByteArray dataOut )
{
	// associate m_pFrameRGB with dataOut
	avpicture_fill( ( AVPicture* )m_pFrameRGB,
		dataOut,
		PIX_FMT_RGB24,
		m_pCodecContext->width, m_pCodecContext->height );

	bool succeeded;

	// check seek flag and seek to the desired frame index
	if( m_bSeeking )
	{
		succeeded = seek( m_iNextFrameIndex );
		if( succeeded )
		{
			succeeded = readNextFrameRaw();
			if( succeeded )
			{
				if( m_pFrameRaw->key_frame == 1 )
				{
					convertRawFrameToRGB();					
				}
				else
				{
					// seek backwards until I see a keyframe

					// TODO: wrap av_seek_frame to return bool					

					int seekTarget = m_iNextFrameIndex - 1;
					succeeded = seek( seekTarget );
					if( succeeded )
					{
						succeeded = readNextFrameRaw();
						
						while( succeeded &&
							m_pFrameRaw->key_frame == 0 )
						{
							--seekTarget;
							succeeded = seek( seekTarget );
							if( succeeded )
							{
								succeeded = readNextFrameRaw();
							}							
						}

						// we found a keyframe
						// now read forward until I get back to my frame number
						if( succeeded )
						{					
							++seekTarget;
							succeeded = readNextFrameRaw();
							while( succeeded &&
								seekTarget < m_iNextFrameIndex )
							{
								++seekTarget;
								succeeded = readNextFrameRaw();
							}
						}
					}
				}
			}
		}

		m_bSeeking = false;
	}
	else
	{
		succeeded = readNextFrameRaw();
	}

	if( succeeded )
	{
		convertRawFrameToRGB();		
		++m_iNextFrameIndex;
	}

	return succeeded;
}

//////////////////////////////////////////////////////////////////////////
// Private
//////////////////////////////////////////////////////////////////////////

FFMPEGVideo::FFMPEGVideo( AVFormatContext* pFormatContext, int iVideoStreamIndex,
						 AVCodecContext* pCodecContext,
						 AVFrame* pFrameRaw,
						 AVFrame* pFrameRGB,
						 SwsContext* pSWSContext ) :

	m_pFormatContext( pFormatContext ),
	m_iVideoStreamIndex( iVideoStreamIndex ),
	m_pCodecContext( pCodecContext ),
	m_pFrameRaw( pFrameRaw ),
	m_pFrameRGB( pFrameRGB ),
	m_pSWSContext( pSWSContext ),

	m_iWidth( pCodecContext->width ),
	m_iHeight( pCodecContext->height ),

	m_bSeeking( false ),
	m_iNextFrameIndex( 0 )
{
	
}

bool FFMPEGVideo::seek( int64 seekTarget )
{
	int retVal = av_seek_frame( m_pFormatContext, m_iVideoStreamIndex,
		seekTarget, AVSEEK_FLAG_ANY );
	return( retVal > 0 );
}

bool FFMPEGVideo::readNextFrameRaw()
{
	// frameFinished:
	// 0 means no frame could be decompressed (end of stream)
	// 1 means it worked
	int retVal;
	int frameFinished = 0;
	AVPacket packet;
	
	// try reading a frame
	retVal = av_read_frame( m_pFormatContext, &packet );
	// loop while we can still read frames && frame isn't finished
	// && is in this order
	// short circuits if we can't read frames anymore
	while( ( retVal >= 0 ) &&
		( frameFinished == 0 ) )
	{
		// check if the packet is from the video s tream
		if( packet.stream_index == m_iVideoStreamIndex )
		{
			// if so, then decode the video frame
			avcodec_decode_video( m_pCodecContext,
				m_pFrameRaw, &frameFinished, 
				packet.data, packet.size );

			// if we managed to decode the next frame
			if( frameFinished )
			{
				// free the packet allocated by av_read_frame
				av_free_packet( &packet );
				return true;
			}
		}
		
		// free the packet allocated by av_read_frame
		av_free_packet( &packet );
		// and try reading the next frame
		retVal = av_read_frame( m_pFormatContext, &packet );
	}

	return false;
}

void FFMPEGVideo::convertRawFrameToRGB()
{
	sws_scale
	(
		m_pSWSContext, // converter
		m_pFrameRaw->data, m_pFrameRaw->linesize, // source data and stride
		0, m_iHeight, // starting y and height
		m_pFrameRGB->data, m_pFrameRGB->linesize
	);
}

// static
bool FFMPEGVideo::s_bInitialized = false;
