#include "video/ImageDirectoryVideo.h"

#include <cassert>
#include <QDir>
#include <QFileInfo>

#include "math/MathUtils.h"
#include "color/QImageUtils.h"

//////////////////////////////////////////////////////////////////////////
// Public
//////////////////////////////////////////////////////////////////////////

// static
Reference< IVideo > ImageDirectoryVideo::fromDirectory( QString directoryName,
													   int framePeriodMilliseconds )
{
	// check if it's a directory
	QFileInfo qfi( directoryName );
	if( !( qfi.isDir() ) )
	{
		fprintf( stderr, "%s is not a directory.\n", qPrintable( directoryName ) );
		assert( false );
		return NULL;
	}

	// ok, it's a directory, create it and take a look
	// get all the filenames that are images, in sorted order
	QStringList nameFilters;
	nameFilters << "*.png" << "*.bmp" << "*.ppm" << "*.pgm" << "*.jpg";

	QDir directory( directoryName );	
	directory.setSorting( QDir::Name );
	
	QStringList filenames = directory.entryList( nameFilters, QDir::Files );
	if( filenames.size() < 1 )
	{
		fprintf( stderr, "No valid images found in %s.\n", qPrintable( directoryName ) );
		assert( false );
		return NULL;
	}

	return new ImageDirectoryVideo( directoryName, filenames, framePeriodMilliseconds );
}

// virtual
int64 ImageDirectoryVideo::numFrames()
{
	return m_qslFilenames.size();
}

// virtual
int ImageDirectoryVideo::framePeriodMilliseconds()
{
	return m_iFramePeriodMilliseconds;
}

// virtual
float ImageDirectoryVideo::framePeriodSeconds()
{
	return( m_iFramePeriodMilliseconds / 1000.f );
}

// virtual
int ImageDirectoryVideo::width()
{
	return m_iWidth;
}

// virtual
int ImageDirectoryVideo::height()
{
	return m_iHeight;
}

// virtual
int ImageDirectoryVideo::bytesPerFrame()
{
	return( 3 * width() * height() );
}

// virtual
int64 ImageDirectoryVideo::getNextFrameIndex()
{
	return m_iNextFrameIndex;
}

// virtual
void ImageDirectoryVideo::setNextFrameIndex( int64 frameIndex )
{
	// TODO: if loop()

	m_iNextFrameIndex = static_cast< int >( frameIndex );
	m_iNextFrameIndex = MathUtils::clampToRangeInt
		(
			static_cast< int >( frameIndex ),
			0, numFrames()
		);

	m_qiNextFrame.load( m_qsDirectoryName + "/" + m_qslFilenames[ m_iNextFrameIndex ] );
}

// virtual
bool ImageDirectoryVideo::getNextFrame( UnsignedByteArray dataOut )
{
	QImageUtils::convertQImageToRGBArray( m_qiNextFrame, dataOut );

	setNextFrameIndex( m_iNextFrameIndex + 1 );

	// TODO: fix this
	return true;
}

//////////////////////////////////////////////////////////////////////////
// Private
//////////////////////////////////////////////////////////////////////////

ImageDirectoryVideo::ImageDirectoryVideo( QString directoryName,
										 QStringList filenames,
										 int framePeriodMilliseconds ) :

	m_qsDirectoryName( directoryName ),
	m_qslFilenames( filenames ),
	m_iFramePeriodMilliseconds( framePeriodMilliseconds ),

	m_iNextFrameIndex( 0 )
{
	QString nextFilename = m_qsDirectoryName + "/" + m_qslFilenames[ m_iNextFrameIndex ];
	
	bool succeeded = m_qiNextFrame.load( nextFilename );
	if( !succeeded )
	{
		fprintf( stderr, "Loading %s failed!\n", qPrintable( nextFilename ) );
		assert( false );
	}

	m_iWidth = m_qiNextFrame.width();
	m_iHeight = m_qiNextFrame.height();
}
